#include <RTI.hh>
#include <fedtime.hh>
#include <NullFederateAmbassador.hh>
#include <libhla/MessageBuffer.hh>
#include <memory>
#include <iostream>
#include <cstdlib>
using namespace std;


// VARIABLES GLOBALES



namespace{
    
    RTIfedTime temps(0.0);
    RTIfedTime lookahead(1.0);
    RTIfedTime endTime(10);
    bool timeRegulation = false;
    bool timeConstrained = false;

    int drone_ID;
    int X_ID;

    int calc_ID;
    int H_ID;
}

class Federe_Interactif : public RTI::RTIambassador, public NullFederateAmbassador
{
    double X;

public:
    RTI::ObjectHandle myHandle;
    Federe_Interactif() {X=42;tag=false;}
    virtual ~Federe_Interactif() throw (RTI::FederateInternalError) {}
    bool tag;
private:

  void timeRegulationEnabled(const RTI::FedTime& theTime)
    throw (
      RTI::InvalidFederationTime,
      RTI::EnableTimeRegulationWasNotPending,
      RTI::FederateInternalError)
  {
     std::cout << std::endl << "callback timeRegulationEnabled pour l'instant " <<
               ((RTIfedTime&)theTime).getTime() << std::endl;
     timeRegulation = true;
  }

  void timeConstrainedEnabled(const RTI::FedTime& theTime)
    throw (
      RTI::InvalidFederationTime,
      RTI::EnableTimeConstrainedWasNotPending,
      RTI::FederateInternalError)
  {
     std::cout << std::endl << "callback timeRegulationConstrained pour l'instant " <<
               ((RTIfedTime&)theTime).getTime() << std::endl;
     timeConstrained = true;
  }

 void timeAdvanceGrant (const RTI::FedTime& theTime)
  throw (
    RTI::InvalidFederationTime,
    RTI::TimeAdvanceWasNotInProgress,
    RTI::FederateInternalError)
  {
    temps = theTime;
    tag= true;
    std::cout << std::endl << "t=" << ((RTIfedTime&)theTime).getTime() << " : <----- timeAdvanceGrant(" << ((RTIfedTime&)theTime).getTime() << ")"<< std::endl << std::endl;
  }
 
  void
  reflectAttributeValues(
    RTI::ObjectHandle theObject,
    const RTI::AttributeHandleValuePairSet& theAttributes,
    const RTI::FedTime&,
    const char *,
    RTI::EventRetractionHandle)
    throw (RTI::ObjectNotKnown,
           RTI::AttributeNotKnown,
           RTI::InvalidFederationTime,
           RTI::FederateInternalError)
  {
      libhla::MessageBuffer buffer;
      double val = 0 ;

      RTI::ULong valueLength ;

      for (unsigned int j=0 ; j<theAttributes.size(); j++) {

	  RTI::AttributeHandle parmHandle = theAttributes.getHandle(j);
	  valueLength = theAttributes.getValueLength(j);
	  
	  buffer.resize(valueLength);
	  buffer.reset();
	  
	  theAttributes.getValue(j, static_cast<char*>(buffer(0)), valueLength);
	  buffer.assumeSizeFromReservedBytes();

	  double h = buffer.read_double();
	  std::cout<<"H received"<<h<<std::endl;
	  X =  h+2;
      }
  }
  
  public:
  void send(){
      libhla::MessageBuffer buffer;
      auto_ptr<RTI::AttributeHandleValuePairSet> attributeSet
	  (RTI::AttributeSetFactory::create(1));

   
    buffer.reset();
    buffer.write_double(X);
    buffer.updateReservedBytes();
    attributeSet->add(X_ID, static_cast<char*>(buffer(0)),buffer.size());    

    try {
	RTIfedTime next_step(temps + lookahead);
	updateAttributeValues(myHandle, *attributeSet, next_step, "coucou");
	std::cout<<"UAV done at time"<<temps.getTime()<<std::endl;
    }
    catch (RTI::Exception& e) {
        std::cout<<"Exception "<<e._name<<" ("<<e._reason<<")"<<std::endl;
    }
    
  }
};


void callTICK (void);
void callCFE (const char* const);
void callJFE (char * name);
void callRFE (void);
void callDFE (void);

void callETR(void);
void callETC(void);
void callNER(void);


void PubSub();

//VARIABLE
Federe_Interactif* myFedere_Interactif = new Federe_Interactif();

int main(int argc, char **argv)
{

  int i;
  float d;
  char c, entree;
  string commande;
  std::cout<<argc<<std::endl;
  if (argc!=3)
  {
      cout<<"prog federate_name absolute_path_to_fom"<<std::endl;
      exit(1);
  }

    //CREER
    callCFE (argv[2]);

    //REJOINDRE
    callJFE (argv[1]);

    PubSub();

    //DECLARER REGULATEUR ET CONTRAINT
    callETR ();
    callETC ();


  //MENU PRINCIPAL
    getchar();
  while(temps  < endTime)
  {
      myFedere_Interactif->send();
      callNER();
      while(!myFedere_Interactif->tag){
	  callTICK();
      }
      myFedere_Interactif->tag = false;
  }
  callRFE();
  exit(1);
}

void callTICK (void) {
    myFedere_Interactif->tick(0.1, 0.2);
}

// FEDERARION MANAGEMENT
// createFederationExecution
void callCFE (const char* const path){
    int test = 1;
    try {
       myFedere_Interactif->createFederationExecution("Federation_Interactive",path);
    }
    catch (RTI::Exception& e) {
        test = 0;
        std::cout << "catch " << e._name << " reason " << e._reason <<std::endl;
    }
    if (test)
       std::cout << "federation creee" << std::endl;
    else
       std::cout << "federation non creee" << std::endl;
}
// deleteFederationExecution
void callDFE (void) {
    int test = 1;
    try {
       myFedere_Interactif->destroyFederationExecution("Federation_Interactive");
    }
    catch (RTI::Exception& e) {
        test = 0;
        std::cout << "catch " << e._name << " reason " << e._reason <<std::endl;
    }
    if (test)
       std::cout << "federation detruite" << std::endl;
    else
       std::cout << "federation non detruite" << std::endl;
}
// joinFederationExecution
void callJFE (char *name) {
    int test = 1;
    try {
       myFedere_Interactif->joinFederationExecution(name, "Federation_Interactive", myFedere_Interactif);
    }
    catch (RTI::Exception& e) {
        test = 0;
        std::cout << "catch " << e._name << " reason " << e._reason <<std::endl;
    }
    if (test)
       std::cout << "federation rejointe" << std::endl;
    else
       std::cout << "federation non rejointe" << std::endl;
}
// resignFederationExecution
void callRFE (void) {
    int test = 1;
    try {
       myFedere_Interactif->resignFederationExecution(RTI::DELETE_OBJECTS_AND_RELEASE_ATTRIBUTES);
    }
    catch (RTI::Exception& e) {
        test = 0;
        std::cout << "catch " << e._name << " reason " << e._reason <<std::endl;
    }
    if (test)
       std::cout << "federation quittee" << std::endl;
    else
       std::cout << "federation non quittee" << std::endl;
}


// TIME MANAGEMENT
// enableTimeRegulation
void callETR (void) {
    try {
       myFedere_Interactif->enableTimeRegulation(temps,lookahead);
    }
    catch (RTI::Exception& e) {
        cout << "catch " << e._name << " reason " << e._reason << endl;
    }
    while (!timeRegulation)
    {
        myFedere_Interactif->tick(0.1, 0.2);
    }
    cout << "federe regulateur" << endl;
}
// enableTimeConstrained
void callETC (void) {
    try {
       myFedere_Interactif->enableTimeConstrained();
    }
    catch (RTI::Exception& e) {
        cout << "catch " << e._name << " reason " << e._reason << endl;
    }
    while (!timeConstrained)
    {
        myFedere_Interactif->tick(0.1, 0.2);

    }
    cout << "federe contraint" << endl;
}
// nextEventRequest
void callNER (void) {
    const float d = 100;
    int test = 1;
    try {
       myFedere_Interactif->nextEventRequest(RTIfedTime(d));
    }
    catch (RTI::Exception& e) {
        test = 0;
        cout << "catch " << e._name << " reason " << e._reason << endl;
    }
    if (test)
       cout << endl << "t=" << temps.getTime() << " nextEventRequest(" << d << ")" << endl;
    else
       cout << "nextEventRequest a echoue" << endl;
}

void PubSub(){

    {
	drone_ID = myFedere_Interactif->getObjectClassHandle("quad");
	calc_ID  = myFedere_Interactif->getObjectClassHandle("calcul");

	X_ID =  myFedere_Interactif->getAttributeHandle("X", drone_ID);
	H_ID =  myFedere_Interactif->getAttributeHandle("H", calc_ID);
    }
    
    //Pub 
    {
    auto_ptr<RTI::AttributeHandleSet> attributes
	(RTI::AttributeHandleSetFactory::create(1));    
    attributes->add(X_ID);    
    myFedere_Interactif->publishObjectClass(drone_ID, *attributes);
    myFedere_Interactif->myHandle = myFedere_Interactif->registerObjectInstance(drone_ID,"quad_obj");
    }

    //sub
    {
    auto_ptr<RTI::AttributeHandleSet> attributes
	(RTI::AttributeHandleSetFactory::create(1));
    
    attributes->add(H_ID);    
    myFedere_Interactif->subscribeObjectClassAttributes(calc_ID, *attributes, RTI::RTI_TRUE);
    }
   
}
